/*************************************************************************
* (C) 2013 AITNET ltd - Sofia/Bulgaria - <misho@aitnet.org>
*  by Michael Pounov <misho@elwix.org>
*
* $Author: misho $
* $Id: elwix.c,v 1.7 2015/06/25 17:53:50 misho Exp $
*
**************************************************************************
The ELWIX and AITNET software is distributed under the following
terms:

All of the documentation and software included in the ELWIX and AITNET
Releases is copyrighted by ELWIX - Sofia/Bulgaria <info@elwix.org>

Copyright 2004 - 2015
	by Michael Pounov <misho@elwix.org>.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
This product includes software developed by Michael Pounov <misho@elwix.org>
ELWIX - Embedded LightWeight unIX and its contributors.
4. Neither the name of AITNET nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AITNET AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
#include "global.h"


int elwix_Debug;
int elwix_Verbose;


/* Memory management */

void *(*e_malloc)(size_t) = malloc;
void *(*e_calloc)(size_t, size_t) = calloc;
void *(*e_realloc)(void*, size_t) = realloc;
char *(*e_strdup)(const char*) = strdup;
void (*e_free)(void*) = free;


#pragma GCC visibility push(hidden)

int use_mm;
const char elwix_Prog[STRSIZ];

int elwix_Errno;
char elwix_Error[STRSIZ];

#pragma GCC visibility pop


// elwix_SetProg() Set program memory pool name
void
elwix_SetProg(const char *csProgName)
{
	strlcpy((char*) elwix_Prog, csProgName, sizeof elwix_Prog);
}

// elwix_GetProg() Get program memory pool name
const char *
elwix_GetProg()
{
	return elwix_Prog;
}

// elwix_GetErrno() Get error code of last operation
int
elwix_GetErrno()
{
	return elwix_Errno;
}

// elwix_GetError() Get error text of last operation
const char *
elwix_GetError()
{
	return elwix_Error;
}

// elwix_SetErr() Set error to variables for internal use!!!
void
elwix_SetErr(int eno, char *estr, ...)
{
	va_list lst;

	elwix_Errno = eno;
	memset(elwix_Error, 0, sizeof elwix_Error);
	va_start(lst, estr);
	vsnprintf(elwix_Error, sizeof elwix_Error, estr, lst);
	va_end(lst);
}

// elwix_mm_inuse() Check for memory management model
int
elwix_mm_inuse()
{
	return use_mm & ELWIX_MPOOL;
}


#pragma GCC visibility push(hidden)
// init libelwix routine
__attribute__((constructor)) static void
_elwix_init()
{
	elwixInit(MEMMGR, 0);
}

// fini libelwix routine
__attribute__((destructor)) static void
_elwix_fini()
{
	elwixFini();
}

/*
 * elwixInit() - Init libelwix library memory management
 *
 * @mm = memory management (ELWIX_SYSM or ELWIX_MPOOL)
 * @maxmem = memory limit
 * return: -1 error or !=-1 used memory management model
 */
int
elwixInit(int mm, u_long maxmem)
{
	switch (mm) {
		case ELWIX_MPOOL:	/* mpool */
			elwix_mpool = mpool_init(maxmem);
			if (elwix_mpool) {
				e_malloc = mpool_xmalloc;
				e_calloc = mpool_xcalloc;
				e_realloc = mpool_xrealloc;
				e_strdup = mpool_xstrdup;
				e_free = mpool_xfree;
				break;
			} else {
				mm = ELWIX_SYSM;
				#undef USE_MPOOL
			}
		case ELWIX_SYSM:	/* system */
			e_malloc = malloc;
			e_calloc = calloc;
			e_realloc = realloc;
			e_strdup = strdup;
			e_free = free;
			break;
		default:		/* not supported */
			elwix_SetErr(EINVAL, "Not supported memory management");
			return -1;
	}

	return (use_mm = mm);
}

/*
 * elwixFini() - Finish libelwix library memory management
 *
 * return: none
 */
void
elwixFini()
{
	switch (use_mm) {
		case ELWIX_MPOOL:
			e_malloc = malloc;
			e_calloc = calloc;
			e_realloc = realloc;
			e_strdup = strdup;
			e_free = free;
			use_mm = ELWIX_SYSM;

			mpool_destroy(&elwix_mpool);
			break;
	}
}
#pragma GCC visibility pop

/*
 * elwix_byteOrder() - Detect platform byte order
 *
 * return: 1 = little endian or 0 big endian
 */
int
elwix_byteOrder()
{
	int x = 1;

	return *(char*) &x;
}
