/*************************************************************************
* (C) 2009 AITNET ltd - Sofia/Bulgaria - <misho@aitbg.com>
*  by Michael Pounov <misho@openbsd-bg.org>
*
* $Author: misho $
* $Id: hash.c,v 1.4 2015/06/25 17:53:50 misho Exp $
*
**************************************************************************
The ELWIX and AITNET software is distributed under the following
terms:

All of the documentation and software included in the ELWIX and AITNET
Releases is copyrighted by ELWIX - Sofia/Bulgaria <info@elwix.org>

Copyright 2004 - 2015
	by Michael Pounov <misho@elwix.org>.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
This product includes software developed by Michael Pounov <misho@elwix.org>
ELWIX - Embedded LightWeight unIX and its contributors.
4. Neither the name of AITNET nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AITNET AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
#include "global.h"


/*
 * hash_varchar() - Compute index hash by variable length string
 *
 * @csStr = Input data buffer
 * @nStrLen = Length of data buffer
 * @nVer = Version of algorythm; 0 - original, 1 - AITNET variant
 * return: Hash value
*/
u_int
hash_varchar(const char *csStr, int nStrLen, int nVer)
{
	register u_int n, hash = 0;
	register int i;

	assert(csStr);

	for (i = 0; i < nStrLen; i++) {
		n = 2 * hash + csStr[i];
		if (!nVer) {
			if (hash & 0x80000000)
				n ^= POLY_CRC32;	/* Polynom CRC-32 aviation */
		} else {
			if (n & 1)			/* Misho ;) patch for better avalanche!!! */
				n ^= POLY_CRC32;	/* Polynom CRC-32 aviation */
		}
		hash = n;
	}

	return hash;
}

/*
 * hash_bernstein() - Compute index hash by Bernstein
 *
 * @csStr = Input data buffer
 * @nStrLen = Length of data buffer
 * @nVer = Version of algorythm; 0 - Bernstein, 1 - DJBX33A variant
 * return: Hash value
*/
u_int
hash_bernstein(const char *csStr, int nStrLen, int nVer)
{
	register u_int hash = INIT_BERNSTEIN;
	register int i;

	assert(csStr);

	for (i = 0; i < nStrLen; i++)
		if (!nVer)
			hash = ((hash << 5) + hash) + csStr[i];
		else
			hash = hash * 33 + csStr[i];	/* DJBX33A */

	return hash;
}

/*
 * hash_fnv1() - Compute index hash by FNV-1
 *
 * @csStr = Input data buffer
 * @nStrLen = Length of data buffer
 * @nVer = Version of algorythm; 0 - FNV-1, 1 - FNV-1a (best avalanche)
 * return: Hash value
*/
u_int
hash_fnv1(const char *csStr, int nStrLen, int nVer)
{
	register u_int hash = INIT_FNV1;
	register int i;

	assert(csStr);

	for (i = 0; i < nStrLen; i++)
		if (!nVer) {				/* FNV-1 */
			hash *= POLY_FNV1;
			hash ^= csStr[i];
		} else {				/* FNV-1a */
			hash ^= csStr[i];
			hash *= POLY_FNV1;
		}

	return hash;
}

/*
 * hash_jenkins() - Compute index hash by Jenkins (one-at-a-time)
 *
 * @csStr = Input data buffer
 * @nStrLen = Length of data buffer
 * return: Hash value
*/
u_int
hash_jenkins(const char *csStr, int nStrLen)
{
	register u_int hash = 0;
	register int i;

	assert(csStr);

	for (i = 0; i < nStrLen; i++) {
		hash += csStr[i];
		hash += (hash << 10);
		hash ^= (hash >> 6);
	}
	hash += (hash << 3);
	hash ^= (hash >> 11);
	hash += (hash << 15);

	return hash;
}

/*
 * hash_reddragon() - Compute index hash by Red Dragon
 *
 * @csStr = Input data buffer
 * @nStrLen = Length of data buffer
 * return: Hash value
*/
u_int
hash_reddragon(const char *csStr, int nStrLen)
{
	register u_int g, hash;
	register int i;

	assert(csStr);

	for (hash = 0, i = 0; i < nStrLen; i++) {
		hash = (hash << 4) + csStr[i];
		if ((g = hash & 0xF0000000)) {
			hash ^= g >> 24;
			hash ^= g;
		}
	}

	return hash;
}

/*
 * hash_jenkins32() - Fast Jenkins hash function
 *
 * @buf = Input buffer
 * @len = Length of buffer
 * @prevhash = Previous hash, if participate in continuing hash
 * return: Hash value
 */
u_int
hash_jenkins32(const u_int *buf, int len, u_int prevhash)
{
	register u_int a, b, c;

	assert(buf);

	a = b = c = INIT_JENKINS32 + (((u_int) len) << 2) + prevhash;

	while (len > 3) {
		a += buf[0];
		b += buf[1];
		c += buf[2];
		MIX32(a, b, c);
		len -= 3;
		buf += 3;
	}

	switch (len) {
		case 3:
			c += buf[2];
		case 2:
			b += buf[1];
		case 1:
			a += buf[0];
			FINAL32(a, b, c);
		case 0:
			break;
	}

	return c;
}
