/*************************************************************************
* (C) 2013 AITNET ltd - Sofia/Bulgaria - <misho@aitnet.org>
*  by Michael Pounov <misho@elwix.org>
*
* $Author: misho $
* $Id: net.c,v 1.12 2015/06/25 17:53:50 misho Exp $
*
**************************************************************************
The ELWIX and AITNET software is distributed under the following
terms:

All of the documentation and software included in the ELWIX and AITNET
Releases is copyrighted by ELWIX - Sofia/Bulgaria <info@elwix.org>

Copyright 2004 - 2015
	by Michael Pounov <misho@elwix.org>.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
This product includes software developed by Michael Pounov <misho@elwix.org>
ELWIX - Embedded LightWeight unIX and its contributors.
4. Neither the name of AITNET nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AITNET AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
#include "global.h"


static char hexlist[] = "0123456789abcdef";

/*
 * e_link_ntoa() - String ethernet address from link address
 *
 * @sdl = link address
 * return: =NULL error or !=NULL ethernet address, should be e_free()
 */
char *
e_link_ntoa(const struct sockaddr_dl *sdl)
{
	static char obuf[64];
	char *out = obuf;
	int i;
	u_char *in = (u_char*) LLADDR(sdl);
	u_char *inlim = in + sdl->sdl_alen;
	int firsttime = 1;

	if (sdl->sdl_nlen) {
		memcpy(obuf, sdl->sdl_data, sdl->sdl_nlen);
		out += sdl->sdl_nlen;
		if (sdl->sdl_alen)
			*out++ = '!';
	}

	while (in < inlim) {
		if (firsttime)
			firsttime ^= firsttime;
		else
			*out++ = ':';

		i = *in++;
		if (i > 0xf) {
			out[1] = hexlist[i & 0xf];
			i >>= 4;
		} else {
			out[1] = hexlist[i];
			i = 0;
		}

		out[0] = hexlist[i];
		out += 2;
	}

	*out = 0;
	return obuf;
}

/*
 * e_link_addr() - String ethernet address to link address
 *
 * @mac = ethernet address
 * @sdl = link address
 * return: -1 error or 0 ok
 */
int
e_link_addr(const char *mac, struct sockaddr_dl * __restrict sdl)
{
	if (!mac || !sdl)
		return -1;
	if (!sdl->sdl_len)
		sdl->sdl_len = sizeof(struct sockaddr_dl);

	link_addr(mac, sdl);
	return 0;
}

/*
 * e_ether_ntoa() - Convert ethernet address to string
 *
 * @n = ethernet address structure, like struct ether_addr
 * @a = string
 * @len = string length
 * return: NULL error or !=NULL string a
 */
char *
e_ether_ntoa(const ether_addr_t * __restrict n, char * __restrict a, int len)
{
	if (!n || !a)
		return NULL;

	memset(a, 0, len);
	if (snprintf(a, len, "%02hhx:%02hhx:%02hhx:%02hhx:%02hhx:%02hhx", 
			n->octet[0], n->octet[1], 
			n->octet[2], n->octet[3], 
			n->octet[4], n->octet[5]) < 17)
		return NULL;

	return a;
}

/*
 * e_ether_aton() - Convert string to ethernet address
 *
 * @a = string
 * @e = ethernet address structure, like struct ether_addr
 * return: NULL error or !=NULL ethernet address structure
 */
ether_addr_t *
e_ether_aton(const char *a, ether_addr_t * __restrict e)
{                       
	int i;

	if (!a || !e)
		return NULL;

	i = sscanf(a, "%hhx:%hhx:%hhx:%hhx:%hhx:%hhx", 
			&e->octet[0], 
			&e->octet[1], 
			&e->octet[2], 
			&e->octet[3], 
			&e->octet[4], 
			&e->octet[5]);
	if (i != 6)
		return NULL;

	return e;
}

/*
 * e_n2port() - Extract port from network structure
 *
 * @addr = Address
 * return: 0 not supported family type or port number
 */
u_short
e_n2port(sockaddr_t * __restrict addr)
{
	u_short port = 0;

	if (!addr)
		return port;

	switch (addr->sa.sa_family) {
		case AF_INET:
			return ntohs(addr->sin.sin_port);
		case AF_INET6:
			return ntohs(addr->sin6.sin6_port);
		default:
			break;
	}

	return port;
}

/*
 * e_n2addr() - Extract address from network structure
 *
 * @addr = Address
 * @val = Value for store string address
 * return: NULL error or !=NULL string address from val
 */
const char *
e_n2addr(sockaddr_t * __restrict addr, ait_val_t * __restrict val)
{
	char *s, str[INET6_ADDRSTRLEN] = { 0 };
	const char *ret = NULL;

	if (!addr || !val)
		return ret;

	AIT_INIT_VAL(val);
	switch (addr->sa.sa_family) {
		case AF_INET:
			if (!inet_ntop(AF_INET, &addr->sin.sin_addr, str, INET_ADDRSTRLEN)) {
				LOGERR;
				return ret;
			} else
				ret = str;
			break;
		case AF_INET6:
			if (!inet_ntop(AF_INET6, &addr->sin6.sin6_addr, str, INET6_ADDRSTRLEN)) {
				LOGERR;
				return ret;
			} else
				ret = str;
			break;
		case AF_LOCAL:
			ret = addr->sun.sun_path;
			break;
		case AF_LINK:
			if (!(s = e_link_ntoa(&addr->sdl))) {
				LOGERR;
				return ret;
			} else
				ret = s;
			break;
		default:
			elwix_SetErr(EPROTONOSUPPORT, "Unsuported address family %d", 
					addr->sa.sa_family);
			return ret;
	}

	AIT_SET_STR(val, ret);
	return (const char*) AIT_GET_STR(val);
}

/*
 * e_gethostbyname() - Get host and port and make network structure
 *
 * @psHost = Hostname
 * @port = Port
 * @addr = Network address structure
 * return: NULL error or !=NULL network structure
 */
sockaddr_t *
e_gethostbyname(const char *psHost, u_short port, sockaddr_t * __restrict addr)
{
	struct hostent *host = NULL;

	if (!psHost || !addr)
		return NULL;

	if (*psHost != '/') {
		/* resolver */
		host = gethostbyname2(psHost, !strchr(psHost, ':') ? AF_INET : AF_INET6);
		if (!host) {
			elwix_SetErr(EINVAL, "Resolver #%d - %s", h_errno, hstrerror(h_errno));
			return NULL;
		} else {
			memset(addr, 0, sizeof(sockaddr_t));
			addr->sa.sa_family = host->h_addrtype;
		}
	} else {
		memset(addr, 0, sizeof(sockaddr_t));
		addr->sa.sa_family = AF_LOCAL;
	}
		

	switch (addr->sa.sa_family) {
		case AF_INET:
			addr->sin.sin_len = sizeof(struct sockaddr_in);
			addr->sin.sin_family = AF_INET;
			addr->sin.sin_port = htons(port);
			memcpy(&addr->sin.sin_addr, host->h_addr, sizeof addr->sin.sin_addr);
			return addr;
		case AF_INET6:
			addr->sin6.sin6_len = sizeof(struct sockaddr_in6);
			addr->sin6.sin6_family = AF_INET6;
			addr->sin6.sin6_port = htons(port);
			memcpy(&addr->sin6.sin6_addr, host->h_addr, sizeof addr->sin6.sin6_addr);
			return addr;
		case AF_LOCAL:
			addr->sun.sun_len = sizeof(struct sockaddr_un);
			addr->sun.sun_family = AF_LOCAL;
			memset(addr->sun.sun_path, 0, sizeof addr->sun.sun_path);
			snprintf(addr->sun.sun_path, sizeof addr->sun.sun_path, "%s-%hu", psHost, port);
			return addr;
		default:
			elwix_SetErr(EPROTONOSUPPORT, "Unsuported address family %d", addr->sa.sa_family);
			break;
	}

	return NULL;
}

/*
 * e_addrcmp() - Compare network addresses
 *
 * @a = 1st address
 * @b = 2nd address
 * @p = compare and ports, if family is AF_INET or AF_INET6
 * return: 0 is equal or !=0 is different
 */
int
e_addrcmp(sockaddr_t * __restrict a, sockaddr_t * __restrict b, int p)
{
	if (a && b && a->sa.sa_family == b->sa.sa_family)
		switch (a->sa.sa_family) {
			case AF_LOCAL:
				return strcmp(a->sun.sun_path, b->sun.sun_path);
			case AF_INET:
				if (p && (a->sin.sin_port - b->sin.sin_port))
					return (int) !!(a->sin.sin_port - b->sin.sin_port);
				else
					return memcmp(&a->sin.sin_addr, &b->sin.sin_addr, 
							sizeof a->sin.sin_addr);
			case AF_INET6:
				if (p && (a->sin6.sin6_port - b->sin6.sin6_port))
					return (int) !!(a->sin6.sin6_port - b->sin6.sin6_port);
				else
					return memcmp(&a->sin6.sin6_addr, &b->sin6.sin6_addr, 
							sizeof a->sin6.sin6_addr);
			case AF_LINK:
				return memcmp(&a->sdl.sdl_data, &b->sdl.sdl_data, 
						sizeof a->sdl.sdl_data);
			case AF_UNSPEC:
				return memcmp(a, b, sizeof(sockaddr_t));
		}

	return (int) !!(a - b);
}

/*
 * e_usleep() - usleep() replacement for ELWIX
 *
 * @usec = microseconds for sleep
 * return: -1 interrupted by signal or 0 ok
 */
int
e_usleep(u_int usec)
{
	struct timeval tv = { (time_t) (usec / 1000000), (long) (usec % 1000000) };

	return select(0, NULL, NULL, NULL, &tv);
}

/*
 * e_innet() - Test address match in network
 *
 * @net = network
 * @addr = address
 * return: -1 error, 0 match or 1 not match
 */
int
e_innet(netaddr_t * __restrict net, inaddr_t * __restrict addr)
{
	register int i;
	int ret = 0;

	if (!net || !addr)
		return -1;

	switch (net->addr.sa.sa_family) {
		case AF_INET:
			for (i = 0; i < sizeof(struct in_addr); i++) {
				ret = ((caddr_t) &net->addr.sin.sin_addr.s_addr)[i] & 
					net->mask.in.s4_addr[i];
				ret -= addr->in.s4_addr[i] & net->mask.in.s4_addr[i];
				if (ret)
					break;
			}
			break;
		case AF_INET6:
			for (i = 0; i < sizeof(struct in6_addr); i++) {
				ret = net->addr.sin6.sin6_addr.s6_addr[i] & 
					net->mask.in6.s6_addr[i];
				ret -= addr->in6.s6_addr[i] & net->mask.in6.s6_addr[i];
				if (ret)
					break;
			}
			break;
		default:
			return -1;
	}

	return !!ret;
}

/*
 * e_getnet() - Get network from string
 *
 * @net = Network string (format: <net[/cidr]>)
 * return: NULL error or !=NULL network should be e_free()
 */
netaddr_t *
e_getnet(const char *net)
{
	netaddr_t *n;
	char *str, *wrk;
	struct hostent *host;

	n = e_malloc(sizeof(netaddr_t));
	if (!n) {
		LOGERR;
		return NULL;
	} else
		memset(n, 0, sizeof(netaddr_t));
	str = e_strdup(net);
	if (!str) {
		LOGERR;
		e_free(n);
		return NULL;
	}
	wrk = strchr(str, '/');
	if (wrk)
		*wrk++ = 0;

	host = gethostbyname2(str, strchr(str, ':') ? AF_INET6 : AF_INET);
	if (!host) {
		elwix_SetErr(EINVAL, "Resolver #%d - %s", h_errno, hstrerror(h_errno));
		e_free(str);
		e_free(n);
		return NULL;
	}
	switch (host->h_addrtype) {
		case AF_INET:
			n->addr.sin.sin_len = sizeof(struct sockaddr_in);
			n->addr.sin.sin_family = host->h_addrtype;
			memcpy(&n->addr.sin.sin_addr, host->h_addr, sizeof n->addr.sin.sin_addr);
			if (wrk)
				n->mask.in.s_addr = E_CIDRMASK(strtol(wrk, NULL, 10));
			else
				n->mask.in.s_addr = 0xFFFFFFFF;
			break;
		case AF_INET6:
			n->addr.sin6.sin6_len = sizeof(struct sockaddr_in6);
			n->addr.sin6.sin6_family = host->h_addrtype;
			memcpy(&n->addr.sin6.sin6_addr, host->h_addr, sizeof n->addr.sin6.sin6_addr);
			/* TODO: should support ipv6 mask */
			break;
		default:
			elwix_SetErr(EINVAL, "Unsupported family #%d", host->h_addrtype);
			e_free(str);
			e_free(n);
			return NULL;
	}

	e_free(str);
	return n;
}

/*
 * e_ether_addr() - Get or set ethernet address from interface name
 *
 * @ifname = interface name
 * @addr = if addr is !=NULL then set this for new address
 * return: NULL error or !=NULL get ethernet address should be e_free()
 */
ether_addr_t *
e_ether_addr(const char *ifname, ether_addr_t * __restrict addr)
{
	ether_addr_t *a = NULL;
	struct ifaddrs *p, *ifa = NULL;

	if (!ifname)
		return NULL;

	a = e_malloc(sizeof(ether_addr_t));
	if (!a)
		return NULL;
	else
		memset(a, 0, sizeof(ether_addr_t));

	getifaddrs(&ifa);
	for (p = ifa; p && p->ifa_name; p++)
		if (p->ifa_name && !strcmp(p->ifa_name, ifname) && p->ifa_addr && 
				p->ifa_addr->sa_family == AF_LINK) {
			memcpy(a, LLADDR((struct sockaddr_dl*) p->ifa_addr), sizeof(ether_addr_t));
			break;
		}
	freeifaddrs(ifa);

	return a;
}

/*
 * e_get1stiface() - Get first interface of host
 *
 * @szIface = interface string buffer
 * @iflen = size of interface buffer
 * return: -1 error or 0 ok
 */
int
e_get1stiface(char *szIface, int iflen)
{
	struct ifaddrs *ifa;

	if (!szIface || !iflen)
		return -1;

	getifaddrs(&ifa);
	strlcpy(szIface, ifa->ifa_name, iflen);
	freeifaddrs(ifa);
	return 0;
}

/*
 * e_getifacebyname() - Get interface and make network structure
 *
 * @psIface = Interface, if =NULL first interface
 * @addr = Network address structure
 * return: NULL error or !=NULL network structure
 */
sockaddr_t *
e_getifacebyname(const char *psIface, sockaddr_t * __restrict addr)
{
	char szIface[64] = { [0 ... 63] = 0 };
	struct ifaddrs *p, *ifa = NULL;

	if (!addr)
		return NULL;

	memset(addr, 0, sizeof(sockaddr_t));
	getifaddrs(&ifa);
	strlcpy(szIface, psIface ? psIface : ifa->ifa_name, sizeof szIface);
	for (p = ifa; p && p->ifa_name; p++)
		if (p->ifa_name && !strcmp(p->ifa_name, szIface) && p->ifa_addr && 
				p->ifa_addr->sa_family == AF_LINK) {
			memcpy(&addr->sdl, p->ifa_addr, sizeof(struct sockaddr_dl));
			break;
		}
	freeifaddrs(ifa);

	return addr;
}

/*
 * e_getlinkbyname() - Get host ethernet address and make network structure
 *
 * @psHost = Host ethernet address
 * @addr = Network address structure
 * return: NULL error or !=NULL network structure
 */
sockaddr_t *
e_getlinkbyname(const char *psHost, sockaddr_t * __restrict addr)
{
	ait_val_t v;
	sockaddr_t *a = addr;

	if (!psHost || !addr)
		return NULL;
	else
		memset(addr, 0, sizeof(sockaddr_t));

	AIT_INIT_VAL2(&v, string);
	if (!strchr(psHost, '.'))
		AIT_SET_STR(&v, ":");
	AIT_SET_STRCAT(&v, psHost);

	addr->sdl.sdl_len = sizeof(struct sockaddr_dl);
	if (e_link_addr(AIT_GET_STR(&v), &addr->sdl))
		a = NULL;

	AIT_FREE_VAL(&v);
	return a;
}

/*
 * e_getlinkbyether() - Get ethernet address and make network structure
 *
 * @mac = Ethernet address
 * @idx = Interface index
 * @type = Interface type
 * @iface = Interface name
 * @addr = Network address structure
 * return: NULL error or !=NULL network structure
 */
sockaddr_t *
e_getlinkbyether(const ether_addr_t * __restrict mac, u_short idx, u_char type, 
		const char *iface, sockaddr_t * __restrict addr)
{
	sockaddr_t *a = addr;

	if (!addr)
		return NULL;
	else
		memset(addr, 0, sizeof(sockaddr_t));

	addr->sdl.sdl_len = sizeof(struct sockaddr_dl);
	addr->sdl.sdl_family = AF_LINK;
	addr->sdl.sdl_index = idx;
	addr->sdl.sdl_type = type;
	if (iface && *iface) {
		addr->sdl.sdl_nlen = strlen(iface);
		memcpy(addr->sdl.sdl_data, iface, addr->sdl.sdl_nlen);
	}
	addr->sdl.sdl_alen = sizeof(ether_addr_t);
	memcpy(LLADDR(&addr->sdl), mac, addr->sdl.sdl_alen);

	return a;
}
