/*************************************************************************
* (C) 2013 AITNET ltd - Sofia/Bulgaria - <misho@aitnet.org>
*  by Michael Pounov <misho@elwix.org>
*
* $Author: misho $
* $Id: regex.c,v 1.3 2015/06/25 17:53:50 misho Exp $
*
**************************************************************************
The ELWIX and AITNET software is distributed under the following
terms:

All of the documentation and software included in the ELWIX and AITNET
Releases is copyrighted by ELWIX - Sofia/Bulgaria <info@elwix.org>

Copyright 2004 - 2015
	by Michael Pounov <misho@elwix.org>.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
This product includes software developed by Michael Pounov <misho@elwix.org>
ELWIX - Embedded LightWeight unIX and its contributors.
4. Neither the name of AITNET nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AITNET AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
#include "global.h"


/*
 * regex_Verify() - Function for verify data match in regex expression
 *
 * @csRegex = Regulare expression pattern
 * @csData = Data for check and verify
 * @startPos = Return start positions
 * @endPos = Return end positions
 * return: NULL not match or error; !=NULL begin of matched data
 */
const char *
regex_Verify(const char *csRegex, const char *csData, int *startPos, int *endPos)
{
	regex_t re;
	regmatch_t match;
	char szErr[STRSIZ];
	int ret, flg;
	const char *pos;

	if (!csRegex || !csData)
		return NULL;

	if ((ret = regcomp(&re, csRegex, REG_EXTENDED))) {
		regerror(ret, &re, szErr, STRSIZ);
		elwix_SetErr(ret, "%s", szErr);
		regfree(&re);
		return NULL;
	}

	for (ret = flg = 0, pos = csData; !(ret = regexec(&re, pos, 1, &match, flg)); 
			pos += match.rm_eo, flg = REG_NOTBOL) {
		if (startPos)
			*startPos = match.rm_so;
		if (endPos)
			*endPos = match.rm_eo;

		pos += match.rm_so;
		break;
	}

	if (ret) {
		regerror(ret, &re, szErr, STRSIZ);
		elwix_SetErr(ret, "%s", szErr);
		pos = NULL;
	}

	regfree(&re);
	return pos;
}

/*
 * regex_Get() - Function for get data match in regex expression
 *
 * @csRegex = Regulare expression pattern
 * @csData = Data from get
 * @psString = Returned string if match
 * @strLen = Length of string
 * return: 0 not match; >0 count of returned chars
 */
int
regex_Get(const char *csRegex, const char *csData, char * __restrict psString, int strLen)
{
	int sp, ep, len;
	const char *str;

	if (!csRegex || !csData)
		return -1;

	str = regex_Verify(csRegex, csData, &sp, &ep);
	if (!str)
		return 0;

	len = ep - sp;
	if (psString && strLen) {
		memset(psString, 0, strLen);
		strncpy(psString, str, strLen <= len ? strLen - 1 : len);
	}

	return len;
}

/*
 * regex_Replace() - Function for replace data match in regex expression with newdata
 *
 * @csRegex = Regulare expression pattern
 * @csData = Source data
 * @csNew = Data for replace
 * return: NULL not match or error; !=NULL allocated new string, must be e_free after use!
 */
char *
regex_Replace(const char *csRegex, const char *csData, const char *csNew)
{
	int sp, ep, len;
	char *tmp, *str = NULL;

	if (!csRegex || !csData)
		return NULL;

	if (!regex_Verify(csRegex, csData, &sp, &ep))
		return NULL;

	// ___ before match
	len = sp + 1;
	str = e_malloc(len);
	if (!str)
		return NULL;
	else
		strlcpy(str, csData, len);
	// * replace match *
	if (csNew) {
		len += strlen(csNew);
		tmp = e_realloc(str, len);
		if (!tmp) {
			e_free(str);
			return NULL;
		} else {
			str = tmp;
			strlcat(str, csNew, len);
		}
	}
	// after match ___
	len += strlen(csData) - ep;
	tmp = e_realloc(str, len);
	if (!tmp) {
		e_free(str);
		return NULL;
	} else {
		str = tmp;
		strlcat(str, csData + ep, len);
	}

	return str;
}
