/*************************************************************************
* (C) 2011 AITNET ltd - Sofia/Bulgaria - <misho@aitnet.org>
*  by Michael Pounov <misho@elwix.org>
*
* $Author: misho $
* $Id: sarray.c,v 1.4 2015/06/25 17:53:50 misho Exp $
*
**************************************************************************
The ELWIX and AITNET software is distributed under the following
terms:

All of the documentation and software included in the ELWIX and AITNET
Releases is copyrighted by ELWIX - Sofia/Bulgaria <info@elwix.org>

Copyright 2004 - 2015
	by Michael Pounov <misho@elwix.org>.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
This product includes software developed by Michael Pounov <misho@elwix.org>
ELWIX - Embedded LightWeight unIX and its contributors.
4. Neither the name of AITNET nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AITNET AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
#include "global.h"


/*
 * sarr_Init() - Create and initialize dynamic split-order array
 *
 * @numItems = Number of Items
 * @segLen = Length of segment
 * return: NULL error, != NULL allocated memory for array
 */
sarr_t *
sarr_Init(int numItems, int segLen)
{
	sarr_t *arr = NULL;

	if (segLen < 1)
		return NULL;

	arr = e_malloc(sizeof(sarr_t));
	if (!arr)
		return NULL;

	arr->sarr_num = numItems;
	arr->sarr_seg = segLen;
	arr->sarr_siz = numItems / segLen + 1;
	arr->sarr_data = e_calloc(arr->sarr_siz, sizeof(sarr_seg_t));
	if (!arr->sarr_data) {
		e_free(arr);
		return NULL;
	} else
		memset(arr->sarr_data, 0, arr->sarr_siz * sizeof(sarr_seg_t));

	return arr;
}

/*
 * sarr_Destroy() - Free all data in dynamic split-order array and Destroy array
 *
 * @parr = Array
 * return: none
 */
void
sarr_Destroy(sarr_t ** __restrict parr)
{
	register int i;

	if (!parr || !*parr)
		return;

	for (i = 0; i < (*parr)->sarr_siz; i++)
		if ((*parr)->sarr_data[i]) {
			e_free((*parr)->sarr_data[i]);
			(*parr)->sarr_data[i] = NULL;
		}

	if ((*parr)->sarr_data)
		e_free((*parr)->sarr_data);
	e_free(*parr);
	*parr = NULL;
}

/*
 * sarr_Copy() Copy source split array to destination split array
 *
 * @dest = Destination split array, after use free with sarr_Destroy()
 * @src = Source split array
 * return: -1 error; >0 count of destination split array
 */
int
sarr_Copy(sarr_t ** __restrict dest, sarr_t * __restrict src)
{
	if (!dest || !src)
		return -1;

	*dest = sarr_Init(sarr_Size(src), sarr_Seg(src));
	if (!*dest)
		return -1;

	memcpy((*dest)->sarr_data, src->sarr_data, (*dest)->sarr_siz * sizeof(sarr_seg_t));
	return sarr_Size(*dest);
}

/*
 * sarr_Vacuum() - Vacuum dynamic split-order array, empty segments will be freed
 *
 * @arr = Array
 * return: -1 error, >-1 freed segments
 */
int
sarr_Vacuum(sarr_t * __restrict arr)
{
	register int i, j;
	int cx = 0;
	sarr_seg_t seg;

	if (!arr)
		return -1;

	for (i = 0; i < arr->sarr_siz; i++)
		if (arr->sarr_data[i]) {
			for (j = 0, seg = arr->sarr_data[i]; j < arr->sarr_seg; j++)
				if (seg[j])
					break;
			if (j == arr->sarr_seg) {
				e_free(arr->sarr_data[i]);
				arr->sarr_data[i] = NULL;
				cx++;
			}
		}

	return cx;
}

/*
 * sarr_Grow() - Grow/Shrink dynamic split-order array, Use with care when it shrink!!!
 *
 * @arr = Array
 * @newNumItems = Number of Items
 * return: -1 error, 0 ok
 */
int
sarr_Grow(sarr_t * __restrict arr, int newNumItems)
{
	sarr_seg_t *data;
	int seg, n = 0;
	register int i;

	if (!arr)
		return -1;

	arr->sarr_num = newNumItems;
	seg = newNumItems / arr->sarr_seg + 1;
	if (arr->sarr_siz == seg)
		return n;
	if (arr->sarr_siz < seg)
		n = seg - arr->sarr_siz;
	else
		for (i = seg; i < arr->sarr_siz; i++)
			if (arr->sarr_data[i])
				e_free(arr->sarr_data[i]);

	arr->sarr_siz = seg;
	data = e_realloc(arr->sarr_data, arr->sarr_siz * sizeof(sarr_seg_t));
	if (!data)
		return -1;
	else
		arr->sarr_data = data;
	memset(arr->sarr_data + (arr->sarr_siz - n), 0, n * sizeof(sarr_seg_t));

	return 0;
}

/*
 * sarr_Get() - Get element from dynamic split-order array
 *
 * @arr = Array
 * @idx = Index (warning 1st element is at position 1)
 * return: NULL not found, !=NULL element
 */
void *
sarr_Get(sarr_t * __restrict arr, u_int idx)
{
	void *ret = NULL;
	sarr_seg_t seg;

	if (!arr || idx < 1 || arr->sarr_num < idx)
		return ret;

	seg = arr->sarr_data[idx / arr->sarr_seg];
	if (seg)
		ret = seg[idx % arr->sarr_seg];

	return ret;
}

/*
 * sarr_Get2() - Always get element from dynamic split-order array
 *	Function automatic grow array. Good use for Hash tables! 
 *
 * @arr = Array
 * @idx = Index (warning 1st element is at position 1)
 * return: NULL not found, !=NULL element
 */
void *
sarr_Get2(sarr_t * __restrict arr, u_int idx)
{
	if (!arr || idx < 1)
		return NULL;
	if (arr->sarr_num < idx)
		if (sarr_Grow(arr, idx))
			return NULL;
	return sarr_Get(arr, idx);
}

/*
 * sarr_Set() - Set element to dynamic split-order array
 *
 * @arr = Array
 * @idx = Index (warning 1st element is at position 1)
 * @data = Value
 * return: NULL error or empty, !=NULL old value in element
 */
void *
sarr_Set(sarr_t * __restrict arr, u_int idx, void *data)
{
	void *ret = NULL;
	sarr_seg_t seg;
	register int pos;

	if (!arr || idx < 1 || arr->sarr_num < idx)
		return ret;

	seg = arr->sarr_data[idx / arr->sarr_seg];
	if (!seg) {
		seg = e_calloc(arr->sarr_seg, sizeof(void*));
		if (!seg)
			return ret;
		else
			memset(seg, 0, arr->sarr_seg * sizeof(void*));
		arr->sarr_data[idx / arr->sarr_seg] = seg;
	}

	pos = idx % arr->sarr_seg;
	ret = seg[pos];
	seg[pos] = data;

	return ret;
}

/*
 * sarr_sarr2array() - Convert from split-order array to dynamic array
 *
 * @sa = split array
 * @sarrFree = after convert split array !=0 will be destroyed sarray
 * return: NULL error or != NULL new array
 */
array_t *
sarr_sarr2array(sarr_t ** __restrict sa, int sarrFree)
{
	array_t *arr = NULL;
	int el;
	register int i;

	if (!sa || !*sa)
		return NULL;

	el = sarr_Size(*sa);
	arr = array_Init(el);
	if (!arr)
		return NULL;

	for (i = 0; i < el; i++)
		array_Set(arr, i, sarr_Get(*sa, i + 1));

	if (sarrFree) {
		e_free(*sa);
		*sa = NULL;
	}
	return arr;
}

/*
 * sarr_array2sarr() - Convert from dynamic array to split-order array
 *
 * @a = array
 * @segLen = Length of segment
 * @arrFree = after convert array !=0 will be destroyed
 * return: NULL error or != NULL new sarr
 */
sarr_t *
sarr_array2sarr(array_t ** __restrict a, int segLen, int arrFree)
{
	sarr_t *sa = NULL;
	int el;
	register int i;

	if (!a || !*a)
		return NULL;

	el = array_Size(*a);
	sa = sarr_Init(el, segLen);
	if (!sa)
		return NULL;

	for (i = 0; i < el; i++)
		sarr_Set(sa, i + 1, array_Get(*a, i));

	if (arrFree) {
		e_free(*a);
		*a = NULL;
	}
	return sa;
}
